@@ -7,9 +7,11 @@ import xml.etree.ElementTree as ET
 import numpy as np
 import pandas as pd
 import time
+
 import dask
 import dask.array as da
 
+import glob
 import re
 import ESMF
 import pathlib
@@ -31,7 +33,7 @@ class GeoAccessor:
         dxdy = self._obj.coords['dxdy'].values
         return dxdy
 
-    def to_raster(self, var=None, crs_out=None, timefrmtstr='%Y-%m-%dT%H%M%S', _dowork_toraster=None):
+    def to_raster(self, var=None, crs_out=None, timefrmtstr='%Y-%m-%dT%H%M%S', client=None):
         """
         Accessible from the xarray object, e.g., ``df.chm.to_raster(...)``. This allows for converting the entire data array into georeferenced tiffs.
         Will work on every timestep in parallel. Doing so requires dask to be configured to use processes and not threads, as the underlying regridding algorithm is not
@@ -75,15 +77,34 @@ class GeoAccessor:
             var = [var]
 
         work = []
+        # fn = xr.apply_ufunc(_dowork_toraster, self._obj, kwargs={'crs_in': self._obj.rio.crs, 'timefrmtstr': timefrmtstr, 'crs_out': crs_out},dask='allowed')
+        # print(fn)
+        # mapped = xr.map_blocks(_dowork_toraster, self._obj, kwargs={'crs_in': self._obj.rio.crs, 'timefrmtstr': timefrmtstr, 'crs_out': crs_out}, template=self._obj)
+        # mapped.compute()
 
+        if client is not None:
+            print('using fire and forget code path')
+
+        total = 0
         for t in range(0, len(self._obj.time.values)):
             for v in var:
-                df = self._obj.isel(time=t)[v]
+                df = self._obj.isel(time=t)[v].copy()
+                total = total + 1
 
-                task = dask.delayed(_dowork_toraster)(self._obj.rio.crs, timefrmtstr, crs_out, df)
-                work.append(task)
+                if client is None:
+                    task = dask.delayed(_dowork_toraster)(self._obj.rio.crs, timefrmtstr, crs_out, df)
+                    work.append(task)
+                else:
+                    c = client.submit(_dowork_toraster, self._obj.rio.crs, timefrmtstr, crs_out, df)
+                    dask.distributed.fire_and_forget(c)
 
-        dask.compute(*work)
+        if client is None:
+            dask.compute(*work)
+        else:
+            print('Waiting...')
+            while len(glob.glob('*.tif')) != total:
+                print(f"""Found {len(glob.glob('*.tmp'))} of {total}""")
+                time.sleep(5)
 
 
 def _get_shape(mesh, dxdy):
@@ -97,7 +118,7 @@ def _get_shape(mesh, dxdy):
     return int(x/dxdy),int(y/dxdy)
 
 
-@dask.delayed
+# @dask.delayed
 def _load_vtu(fname):
     # print('Loading and combining vtu...',end='')
     vtu = pv.MultiBlock([pv.read(f) for f in fname])
@@ -106,24 +127,13 @@ def _load_vtu(fname):
 
 
 @dask.delayed
-def _regrid_mesh_to_grid(vtu, dxdy, var):
-                         # num_node, nodeId, nodeCoord, nodeOwner,
-                         # num_elem, elemId, elemType, elemConn, elemCoord,
-                         # numX, numY,
-                         # x_center, y_center,
-                         # x_corner, y_corner
-                         # ):
+def _regrid_mesh_to_grid(v, dxdy, var, regridding_method):
 
     print(f'_regrid_mesh_to_grid called for {var}')
-
     start_time = time.time()
+    vtu = _load_vtu(v)
     mesh, grid = _build_regridding_ds(vtu, dxdy)
 
-    # mesh, grid = _regridding_ds_from_partial(num_node, nodeId, nodeCoord, nodeOwner,
-    #                             num_elem, elemId, elemType, elemConn, elemCoord,
-    #                             numX, numY,
-    #                             x_center, y_center,
-    #                             x_corner, y_corner)
     print('Took ' + str(time.time() - start_time) + ' to build the regridding ds')
 
     start_time = time.time()
@@ -137,7 +147,12 @@ def _regrid_mesh_to_grid(vtu, dxdy, var):
     # in the source field is 0.0, then it's possible that after regridding with the second-order method, the
     # destination field will contain values less than 0.0.
     # https://esmf-org.github.io/dev_docs/ESMF_refdoc/node3.html#SECTION03023000000000000000
-    regrid = ESMF.Regrid(srcfield, dstfield, regrid_method=ESMF.RegridMethod.CONSERVE, unmapped_action=ESMF.UnmappedAction.IGNORE)
+
+    method = ESMF.RegridMethod.BILINEAR
+    if regridding_method == 'CONSERVE':
+        method = ESMF.RegridMethod.CONSERVE
+
+    regrid = ESMF.Regrid(srcfield, dstfield, regrid_method=method, unmapped_action=ESMF.UnmappedAction.IGNORE)
     out = regrid(srcfield, dstfield, zero_region=ESMF.Region.SELECT)
 
     df = da.from_array(out.data)
@@ -146,124 +161,7 @@ def _regrid_mesh_to_grid(vtu, dxdy, var):
 
     return df
 
-def _regridding_ds_from_partial(num_node, nodeId, nodeCoord, nodeOwner,
-                                num_elem, elemId, elemType, elemConn, elemCoord,
-                                numX, numY,
-                                x_center, y_center,
-                                x_corner, y_corner):
-    nodes, elements = (0, 1)
-    u, v = (0, 1)
 
-    Emesh = ESMF.Mesh(parametric_dim=2, spatial_dim=2)
-    Emesh.add_nodes(num_node, nodeId, nodeCoord, nodeOwner)
-    Emesh.add_elements(num_elem, elemId, elemType, elemConn, element_coords=elemCoord)
-
-    max_index = np.array([len(x_center), len(y_center)])
-
-    grid = ESMF.Grid(max_index, staggerloc=[ESMF.StaggerLoc.CENTER, ESMF.StaggerLoc.CORNER],
-                     coord_sys=ESMF.CoordSys.CART)
-
-    # RLO: access Grid center coordinates
-    gridXCenter = grid.get_coords(0)
-    gridYCenter = grid.get_coords(1)
-
-    # RLO: set Grid center coordinates as a 2D array (this can also be done 1d)
-    gridXCenter[...] = x_center.reshape((x_center.size, 1))
-    gridYCenter[...] = y_center.reshape((1, y_center.size))
-
-    # RLO: access Grid corner coordinates
-    gridXCorner = grid.get_coords(0, staggerloc=ESMF.StaggerLoc.CORNER)
-    gridYCorner = grid.get_coords(1, staggerloc=ESMF.StaggerLoc.CORNER)
-
-    # RLO: set Grid corner coordinats as a 2D array
-    gridXCorner[...] = x_corner.reshape((x_corner.size, 1))
-    gridYCorner[...] = y_corner.reshape((1, y_corner.size))
-
-    return Emesh, grid
-
-def _build_partial_regridding_ds(mesh, dxdy):
-    print('_build_partial_regridding_ds called')
-
-    nodes, elements = (0, 1)
-    u, v = (0, 1)
-
-    Emesh = ESMF.Mesh(parametric_dim=2, spatial_dim=2)
-    num_node = mesh.n_points
-    num_elem = mesh.n_cells
-
-    nodeId = np.linspace(0, num_node, num_node)
-    nodeCoord = np.empty(num_node * 2)  # bc xy coords
-    nodeOwner = np.zeros(num_node)
-
-    for i in range(0, num_node * 2, 2):
-        # just x, y
-        nodeCoord[i] = mesh.GetPoint(i // 2)[0]
-        nodeCoord[i + 1] = mesh.GetPoint(i // 2)[1]
-
-    Emesh.add_nodes(num_node, nodeId, nodeCoord, nodeOwner)
-
-    elemId = np.empty(num_elem)
-    elemType = np.empty(num_elem)
-    elemConn = np.empty(num_elem * 3)
-    elemCoord = np.empty(num_elem * 2)
-
-    for i in range(0, num_elem * 3, 3):
-        elemId[i // 3] = i
-
-        elemType[i // 3] = ESMF.MeshElemType.TRI
-
-        v0 = mesh.GetCell(i // 3).GetPointId(0)
-        v1 = mesh.GetCell(i // 3).GetPointId(1)
-        v2 = mesh.GetCell(i // 3).GetPointId(2)
-
-        elemConn[i] = v0
-        elemConn[i + 1] = v2
-        elemConn[i + 2] = v1
-
-        centreX = 1 / 3 * (nodeCoord[v0] + nodeCoord[v1] + nodeCoord[v2])
-        centreY = 1 / 3 * (nodeCoord[v0 + 1] + nodeCoord[v1 + 1] + nodeCoord[v2 + 1])
-
-        elemCoord[i // 3] = centreX
-        elemCoord[(i // 3) + 1] = centreY
-
-    Emesh.add_elements(num_elem, elemId, elemType, elemConn, element_coords=elemCoord)
-
-    numX, numY = _get_shape(mesh, dxdy)
-
-    # cell centres
-    x_center = np.linspace(start=Emesh.coords[nodes][u].min() + dxdy / 2.,
-                           stop=Emesh.coords[nodes][u].max() - dxdy / 2., num=numX)
-    y_center = np.linspace(start=Emesh.coords[nodes][v].min() + dxdy / 2.,
-                           stop=Emesh.coords[nodes][v].max() - dxdy / 2., num=numY)
-
-    # node coords
-    x_corner = np.linspace(start=Emesh.coords[nodes][u].min(), stop=Emesh.coords[nodes][u].max(), num=numX + 1)
-    y_corner = np.linspace(start=Emesh.coords[nodes][v].min(), stop=Emesh.coords[nodes][v].max(), num=numY + 1)
-
-    # max_index = np.array([len(x_center), len(y_center)])
-
-    # grid = ESMF.Grid(max_index, staggerloc=[ESMF.StaggerLoc.CENTER, ESMF.StaggerLoc.CORNER],
-    #                  coord_sys=ESMF.CoordSys.CART)
-    #
-    # # RLO: access Grid center coordinates
-    # gridXCenter = grid.get_coords(0)
-    # gridYCenter = grid.get_coords(1)
-    #
-    # # RLO: set Grid center coordinates as a 2D array (this can also be done 1d)
-    # gridXCenter[...] = x_center.reshape((x_center.size, 1))
-    # gridYCenter[...] = y_center.reshape((1, y_center.size))
-    #
-    # # RLO: access Grid corner coordinates
-    # gridXCorner = grid.get_coords(0, staggerloc=ESMF.StaggerLoc.CORNER)
-    # gridYCorner = grid.get_coords(1, staggerloc=ESMF.StaggerLoc.CORNER)
-    #
-    # # RLO: set Grid corner coordinats as a 2D array
-    # gridXCorner[...] = x_corner.reshape((x_corner.size, 1))
-    # gridYCorner[...] = y_corner.reshape((1, y_corner.size))
-
-    return num_node, nodeId, nodeCoord, nodeOwner, num_elem, elemId, elemType, elemConn, elemCoord, numX, numY, x_center, y_center, x_corner, y_corner
-
-# @dask.delayed
 def _build_regridding_ds(mesh, dxdy):
     # print('_build_regridding_ds called')
 
@@ -275,14 +173,21 @@ def _build_regridding_ds(mesh, dxdy):
     num_elem = mesh.n_cells
 
     nodeId = np.linspace(0, num_node, num_node)
-    nodeCoord = np.empty(num_node * 2)  # bc xy coords
+
+    # store as x,y pairs in sequence
+    # [ x0,y0,
+    #   x1,y1. [...] ]
+    nodeCoord = np.empty(num_node * 2)
     nodeOwner = np.zeros(num_node)
 
+
     for i in range(0, num_node * 2, 2):
         # just x, y
         nodeCoord[i] = mesh.GetPoint(i // 2)[0]
         nodeCoord[i + 1] = mesh.GetPoint(i // 2)[1]
 
+
+
     Emesh.add_nodes(num_node, nodeId, nodeCoord, nodeOwner)
 
     elemId = np.empty(num_elem)
@@ -291,7 +196,7 @@ def _build_regridding_ds(mesh, dxdy):
     elemCoord = np.empty(num_elem * 2)
 
     for i in range(0, num_elem * 3, 3):
-        elemId[i // 3] = i
+        elemId[i // 3] = i // 3
 
         elemType[i // 3] = ESMF.MeshElemType.TRI
 
@@ -299,15 +204,25 @@ def _build_regridding_ds(mesh, dxdy):
         v1 = mesh.GetCell(i // 3).GetPointId(1)
         v2 = mesh.GetCell(i // 3).GetPointId(2)
 
+        # nodes that make up the triangulation
         elemConn[i] = v0
         elemConn[i + 1] = v2
         elemConn[i + 2] = v1
 
-        centreX = 1 / 3 * (nodeCoord[v0] + nodeCoord[v1] + nodeCoord[v2])
-        centreY = 1 / 3 * (nodeCoord[v0 + 1] + nodeCoord[v1 + 1] + nodeCoord[v2 + 1])
+        # Centroid
+        # i = x + width*y;
+        centreX = 1 / 3 * (nodeCoord[2*v0 + 0] + nodeCoord[2*v1 + 0] + nodeCoord[2*v2 + 0])
+        centreY = 1 / 3 * (nodeCoord[2*v0 + 1] + nodeCoord[2*v1 + 1] + nodeCoord[2*v2 + 1])
 
-        elemCoord[i // 3] = centreX
-        elemCoord[(i // 3) + 1] = centreY
+        #i = x + width*y;
+        elemCoord[2*(i // 3) + 0] = centreX
+        elemCoord[2*(i // 3) + 1] = centreY
+
+        # if i // 3 < 3:
+        #     print(f'tri={i // 3}')
+        #     print(f'\t\t', nodeCoord[v0*2], nodeCoord[v1*2], nodeCoord[v2*2])
+        #     print(f'\t\t', nodeCoord[2*v0 +1], nodeCoord[2*v1+1], nodeCoord[2*v2+1])
+        #     print('')
 
     Emesh.add_elements(num_elem, elemId, elemType, elemConn, element_coords=elemCoord)
 
@@ -406,7 +321,7 @@ def vtu_to_xarray(fname, dxdy=30, variables=None):
 
     for var, arrays in delayed_vtu.items():
         delayed_vtu[var] = xr.concat(arrays, dim=times)
-        delayed_vtu[var].chunk({'time': 1})
+        delayed_vtu[var] = delayed_vtu[var].chunk({'time': 1})
 
     ds = xr.Dataset(data_vars=delayed_vtu)
     ds = ds.rio.set_crs(proj4)
@@ -433,7 +348,7 @@ def vtu_to_xarray(fname, dxdy=30, variables=None):
 
     return ds
 
-def pvd_to_xarray(fname, dxdy=50, variables=None):
+def pvd_to_xarray(fname, dxdy=50, variables=None, regridding_method='BILINEAR'):
     """
     Opens a pvd file and returns a Dask delayed xarray object. As it is a delayed object,
     once a specific timestep's variable is operated upon, the delayed object is computed
@@ -442,7 +357,8 @@ def pvd_to_xarray(fname, dxdy=50, variables=None):
 
     :param fname: pvd file path
     :param dxdy: spatial resolution in metres to regrid to.
-    :para variables: List of variables to keep from the loaded vtu file
+    :param variables: List of variables to keep from the loaded vtu file
+    :param regridding_method 'BILINEAR' or 'CONSERVE'
     :return:
     """
 
@@ -553,14 +469,9 @@ def pvd_to_xarray(fname, dxdy=50, variables=None):
         # vtu = _load_vtu(v)
 
         for var in variables:
-            print(f'Loading delayed {var}')
-            vtu = _load_vtu(v)  # can done here to pickle, but then will end up being a bit more costly
-            df = _regrid_mesh_to_grid(vtu, dxdy, var)
-                                        # num_node, nodeId, nodeCoord, nodeOwner,
-                                        #  num_elem, elemId, elemType, elemConn, elemCoord,
-                                        #  numX, numY,
-                                        #  x_center, y_center,
-                                        #  x_corner, y_corner)
+            vtu = v #_load_vtu(v)  # can done here to pickle, but then will end up being a bit more costly
+            df = _regrid_mesh_to_grid(vtu, dxdy, var, regridding_method)
+
             d = da.from_delayed(df,
                                 shape=shape,
                                 dtype=np.dtype('float64'))
@@ -575,11 +486,11 @@ def pvd_to_xarray(fname, dxdy=50, variables=None):
 
     _dt = int(dt.astype("timedelta64[s]")/np.timedelta64(1, 's'))
 
-    times = pd.date_range(start = epoch, end = end_time, freq= f'{_dt} s' , name="time")
+    times = pd.date_range(start=epoch, end=end_time, freq=f'{_dt} s', name="time")
 
     for var, arrays in delayed_vtu.items():
         delayed_vtu[var] = xr.concat(arrays, dim=times)
-        delayed_vtu[var].chunk({'time': 1, 'x':-1, 'y':-1})
+        delayed_vtu[var] = delayed_vtu[var].chunk({'time': 1, 'x': -1, 'y': -1})
 
     ds = xr.Dataset(data_vars=delayed_vtu)
 
